--[[
Bison - a framework for free positioning of buffs.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:
	
Free Software Foundation, I.,
51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
--]]

Bison = LibStub( 'AceAddon-3.0'):NewAddon( 'Bison', 'AceEvent-3.0', 'AceConsole-3.0', 'LibDebugLog-1.0')
if not Bison then return end -- already loaded and no upgrade necessary

local L      = LibStub( 'AceLocale-3.0'):GetLocale( 'Bison')
local AceGUI = LibStub( 'AceGUI-3.0')

Bison.name, Bison.localizedname = GetAddOnInfo( 'Bison')
Bison.version = format( '%s: %s |cFF20FF20(Cata)|r', L.Version, GetAddOnMetadata( 'Bison', 'Version') or 'unknown')

local DEFAULTS = { 
	profile = {
		["*"]  = { xPos = 400, yPos = 300, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 16, number = 16, show = true,  timer = true, flashing = true, horizontal = true, sort = 'dec',  style = { 'Blizzard', 0, nil }, },
		buff   = { xPos = 400, yPos = 500, xPadding = 5, yPadding = 15, scale = 1.0, rows = 2, cols = 20, number = 40, show = true,  timer = true, flashing = true, horizontal = true, sort = 'dec',  style = { 'Blizzard', 0, nil }, },
		prog   = { xPos = 400, yPos = 150, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 16, number = 16, show = false, timer = true, flashing = true, horizontal = true, sort = 'dec',  style = { 'Blizzard', 0, nil }, },
		debuff = { xPos = 400, yPos = 250, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 16, number = 16, show = true,  timer = true, flashing = true, horizontal = true, sort = 'dec',  style = { 'Blizzard', 0, nil }, },
		weapon = { xPos = 400, yPos = 200, xPadding = 5, yPadding = 15, scale = 1.0, rows = 1, cols = 2,  number = 2,  show = true,  timer = true, flashing = true, horizontal = true, sort = 'none', style = { 'Blizzard', 0, nil }, },
		join    = false,
		locked  = true,
		enabled = true,
	} 
}
local HIDE_BLIZZ_BARS = true -- for tests false
local SORT_DESC = {
	[1] = L.SortNone,
	[2] = L.SortAlpha,
	[3] = L.SortRevert,
	[4] = L.SortInc,
	[5] = L.SortDec,
	[6] = L.SortDurationAsc,
	[7] = L.SortDurationDesc
}
local SORT_TYPES = {
	['none']   = 1,
	['alpha']  = 2,
	['revert'] = 3,
	['inc']    = 4,
	['dec']    = 5,
	['durationasc']  = 6,
	['durationdesc'] = 7,
}
local oldConsolidate

--[[---------------------------------------------------------------------------------
Local
------------------------------------------------------------------------------------]]
local function SortNone( c1, c2)
	return c1.id < c2.id
end

local function SortNameAsc( c1, c2)
	local a = c1.userdata
	local b = c2.userdata
	if a.name and b.name then return a.name < b.name end
	if a.name then return true end
	if b.name then return false end
	return c1.id < c2.id
end

local function SortNameDesc( c1, c2)
	local a = c1.userdata
	local b = c2.userdata
	if a.name and b.name then return a.name > b.name end
	if a.name then return true end
	if b.name then return false end
	return c1.id < c2.id
end

local function SortTimeLeftAsc( c1, c2)
	local a = c1.userdata
	local b = c2.userdata
	if a.name and b.name then
		if a.timeLeft and b.timeLeft then return a.timeLeft < b.timeLeft end
		if a.timeLeft then return true end
		if b.timeLeft then return false end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return c1.id < c2.id
end

local function SortTimeLeftDesc( c1, c2)
	local a = c1.userdata
	local b = c2.userdata
	if a.name and b.name then
		if a.timeLeft and b.timeLeft then return a.timeLeft > b.timeLeft end
		if a.timeLeft then return false end
		if b.timeLeft then return true end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return c1.id < c2.id
end

local function SortDurationAsc( c1, c2)
	local a = c1.userdata
	local b = c2.userdata
	if a.name and b.name then
		if a.duration and b.duration then return a.duration < b.duration end
		if a.duration then return true end
		if b.duration then return false end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return c1.id < c2.id
end

local function SortDurationDesc( c1, c2)
	local a = c1.userdata
	local b = c2.userdata
	if a.name and b.name then
		if a.duration and b.duration then return a.duration > b.duration end
		if a.duration then return true end
		if b.duration then return false end
		return a.name < b.name
	end
	if a.name then return true end
	if b.name then return false end
	return c1.id < c2.id
end

local SORT = {
	['none']   = SortNone,
	['alpha']  = SortNameAsc,
	['revert'] = SortNameDesc,
	['inc']    = SortTimeLeftAsc,
	['dec']    = SortTimeLeftDesc,
	['durationasc']  = SortDurationAsc,
	['durationdesc'] = SortDurationDesc,
}

--[[---------------------------------------------------------------------------------
Class
------------------------------------------------------------------------------------]]
function Bison:OnInitialize()
	self.db = LibStub( 'AceDB-3.0'):New( 'BisonDB', DEFAULTS, 'Default')
	self.db.RegisterCallback( self, 'OnNewProfile')
	self.db.RegisterCallback( self, 'OnProfileChanged')
	self.db.RegisterCallback( self, 'OnProfileCopied')
	self.db.RegisterCallback( self, 'OnProfileReset')
	self.db.RegisterCallback( self, 'OnProfileDeleted')
end

function Bison:OnEnable()
	self:ToggleDebugLog( false)
	self:HideBlizzardFrame()
	self:Debug( 'Bison enabled')
	self:Print( self.version, 'loaded')
end

function Bison:OnDisable()
	self:Debug( 'Bison disabled')
	self:RestoreBlizzardFrame()
	self:ShowBlizzardFrame()
end

--[[---------------------------------------------------------------------------------
Main
------------------------------------------------------------------------------------]]
function Bison:HideBlizzardFrame()
--	oldConsolidate = _G['CONSOLIDATE_BUFFS']
--	BlizzardOptionsPanel_SetCVarSafe( 'consolidateBuffs', '0')
--	CONSOLIDATE_BUFFS = '0'
--	_G['CONSOLIDATE_BUFFS'] = '0'
	if HIDE_BLIZZ_BARS then
		BuffFrame_Update()
		BuffFrame:Hide()
		TemporaryEnchantFrame:Hide()
		ConsolidatedBuffs:Hide()
	end
end

function Bison:ShowBlizzardFrame()
--	BlizzardOptionsPanel_SetCVarSafe( 'consolidateBuffs', oldConsolidate)
--	CONSOLIDATE_BUFFS = oldConsolidate
--	_G['CONSOLIDATE_BUFFS'] = oldConsolidate
	BuffFrame:Show()
	TemporaryEnchantFrame:Show()
	BuffFrame_Update()
end

function Bison:MoveBlizzardFrame( frame)
	if HIDE_BLIZZ_BARS then
		BuffFrame:ClearAllPoints()
		BuffFrame:SetPoint( 'TOPLEFT', frame, 'BOTTOMLEFT', 0, 0)
		BuffFrame:SetBackdrop{
			bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background', tile = true, tileSize = 16,
			edgeFile = 'Interface\\Tooltips\\UI-Tooltip-Border', edgeSize = 16,
			insets = { left = 3, right = 3, top = 3, bottom = 3 },
		}
		BuffFrame:SetBackdropColor( TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
	end
end

function Bison:RestoreBlizzardFrame()
	if HIDE_BLIZZ_BARS then
		BuffFrame:ClearAllPoints()
		BuffFrame:SetPoint( 'TOPRIGHT', parent, 'TOPRIGHT', -205, -13)
		BuffFrame:SetBackdrop( nil)
	end
end

local function round( value)
	return floor( value + 0.5)
end

local prototype = {
	['DoEnable'] = function( self)
		if self.profile.show then
			self:SetPos()
			self:SetLock()
			self:Refresh()
			self:SetVisible( true)
			self:RegisterMessage( 'BISON_UPDATE')
			self:RegisterMessage( 'BISON_DEBUG')
		end
	end,
	
	['DoDisable'] = function( self)
		self:SetVisible( false)
	end,

	['BISON_DEBUG'] = function( self, event, debug)
		self:ToggleDebugLog( debug)
	end,

	['BISON_UPDATE'] = function( self, event, bar)
		Bison:Debug( event, bar)
		if bar == nil or bar == self.moduleName then
			if self.profile.show then
				self:SetPos()
				self:SetLock()
				self:Refresh()
				self:SetVisible( true)
			else
				self:SetVisible( false)
			end
		end
	end,

	['UNIT_AURA'] = function( self, event, unit)
		if unit == PlayerFrame.unit then
			self:Refresh()
		end
	end,

	['SetVisible'] = function( self, visible)
		if visible and self.profile.show then
			self.group:SetVisible( true)
			self:RegisterEvent( 'UNIT_AURA')
		else
			self.group:SetVisible( false)
			self:UnregisterEvent( 'UNIT_AURA')
		end
	end,

	['SetLock'] = function( self)
		self.group:SetLock( Bison.db.profile.locked)
	end,
	
	['IsLock'] = function( self)
		return Bison.db.profile.locked
	end,
	
	['SetPos'] = function( self)
		local profile = self.profile
		if profile.show then
			self.group:SetPos( profile.xPos, profile.yPos)
		end
	end,
	
	['GetSort'] = function( self, sort)
		return SORT[sort] or SortNone
	end,

	['Refresh'] = function( self)
		self:ReadValues()
		self:UpdateLeft()
		local flashing = self.profile.flashing
		for _,child in pairs( self.group.children) do
			child:UpdateAll( flashing)
		end
		table.sort( self.group.children, self:GetSort( self.profile.sort))
		self.group:DoLayout()
	end,

	['CreateGroup'] = function( self)
		local group = AceGUI:Create( 'BisonGroup')
		group:SetLayout( 'BuffBar')
		self.group = group
		return group
	end,

	['DurationString'] = function( self, seconds, timer)
		if not timer then
			return string.format( SecondsToTimeAbbrev( seconds))
		end
		if not seconds then
			return ''
		end
		local negative, time
		if seconds < 0 then 
			negative = '-'
			time = -seconds
		else
			negative = ''
			time = seconds
		end
		if time < 60 then 
			return string.format( '|cffffffff%s%ds|r', negative, time) 
		end
		local min = floor( time / 60)
		if time < 600 then 
			return string.format( '|cffffffff%s%d:%02d|r', negative, min, time % 60) 
		end
		if time < 3600 then 
			return string.format( '|cffffffff%s%dm|r', negative, min) 
		end
		return string.format( '|cffffffff%s%d:%02dh|r', negative, floor( min / 60), min % 60)
	end,

	['GetProperty'] = function( self, info)
		local key = info[#info]
		return self.profile[key]
	end,
	
	['SetProperty'] = function( self, info, value)
		local key = info[#info]
		self.profile[key] = value
		self:SendMessage( 'BISON_UPDATE', self.moduleName)
	end,
	
	['SetScale'] = function( self, info, value)
		value = tonumber(value)
		if value and value > 0 and value <= 2 then
			self.profile.scale = value
			self:SendMessage( 'BISON_UPDATE', self.moduleName)
		end
	end,
	
	['SetCols'] = function( self, info, value)
		local maxButton = info.option.max or 16
		value = tonumber(value)
		if value and value > 0 and value <= maxButton then
			value = floor( value)
			self.profile.cols = value
			self.profile.rows = ceil( self.profile.number / value)
			self:SendMessage( 'BISON_UPDATE', self.moduleName)
		end
	end,
	
	['SetRows'] = function( self, info, value)
		local maxButton = info.option.max or 16
		value = tonumber(value)
		if value and value > 0 and value <= maxButton then
			value = floor( value)
			self.profile.cols = ceil( self.profile.number / value)
			self.profile.rows = value
			self:SendMessage( 'BISON_UPDATE', self.moduleName)
		end
	end,
	
	['SetNumber'] = function( self, info, value)
		local maxButton = info.option.max or 16
		value = tonumber(value)
		if value and value > 0 and value <= maxButton then
			value = floor( value)
			self.profile.number = floor( value)
			self.profile.cols   = ceil( self.profile.number / self.profile.rows)
			self.profile.rows   = ceil( self.profile.number / self.profile.cols)
			self:SendMessage( 'BISON_UPDATE', self.moduleName)
		end
	end,
	
	['SetXPadding'] = function( self, info, value)
		local min = info.option.mix or -20
		local max = info.option.max or 20
		value = tonumber(value)
		if value and value >= min and value <= max then
			self.profile.xPadding = floor( value)
			self:SendMessage( 'BISON_UPDATE', self.moduleName)
		end
	end,
	
	['SetYPadding'] = function( self, info, value)
		local min = info.option.mix or -50
		local max = info.option.max or 50
		value = tonumber(value)
		if value and value >= min and value <= max then
			self.profile.yPadding = floor( value)
			self:SendMessage( 'BISON_UPDATE', self.moduleName)
		end
	end,
	
	['GetJoin'] = function( self)
		return Bison.db.profile.join
	end,
	
	['SetJoin'] = function( self, info, value)
		Bison.db.profile.join = value
		self:SendMessage( 'BISON_UPDATE', self.moduleName)
	end,
	
	['GetSortType'] = function( self)
		return SORT_TYPES[self.profile.sort]
	end,
	
	['GetSortDesc'] = function( self)
		return SORT_DESC
	end,
	
	['SetSortType'] = function( self, info, value)
		value = tonumber(value)
		if value then
			for k,v in pairs(SORT_TYPES) do
				if v == value then
					self.profile.sort = k
					self:SendMessage( 'BISON_UPDATE', self.moduleName)
					break
				end
			end
		end
	end
}

function Bison:NewBarModule( name)
	if self:GetModule( name, true) then
		error( 'can not register double module: '..name, 2)
	end
	return self:NewModule( name, prototype, 'AceEvent-3.0', 'AceConsole-3.0', 'LibDebugLog-1.0')
end

function Bison:SetDebug( debug)
	self:ToggleDebugLog( debug)
	self:SendMessage( 'BISON_DEBUG', debug)
end

--[[---------------------------------------------------------------------------------
LBF
------------------------------------------------------------------------------------]]
local LBF = LibStub( 'LibButtonFacade', true)

local function LBF_OnSkin( bar, skin, gloss, backdrop, group, button, colors)
	Bison.db.profile[bar].style = { skin, gloss, backdrop, colors } 
end

function Bison:AddToLBF( name, bar)
	if LBF then
		LBF:Group( self.localizedname, name):Skin( unpack( self.db.profile[bar].style))
		LBF:RegisterSkinCallback( self.localizedname, LBF_OnSkin, bar)
	end
end

function Bison:AddChieldToLBF( name, chield)
	if LBF then
		LBF:Group( self.localizedname, name):AddButton( chield.frame)
	end
end

function Bison:RemoveChieldFromLBF( name, chield)
	if LBF then
		LBF:Group( self.localizedname, name):RemoveButton( chield.frame)
	end
end

--[[---------------------------------------------------------------------------------
Profile
------------------------------------------------------------------------------------]]
function Bison:ListProfiles()
	self:Print( L.AvailableProfiles)
	local current = self.db:GetCurrentProfile()
	for _,name in ipairs(self.db:GetProfiles()) do
		self:Print( format( (name == current) and ' - |cFFFFFF00%s|r' or ' - %s', name))
	end
end

function Bison:OnNewProfile( event, db, name)
	self:Update()
	self:Print( format( L.ProfileCreated, name))
end

function Bison:OnProfileChanged( event, db, name)
	self:Update()
	self:Print( format( L.ProfileLoaded, name))
end

function Bison:OnProfileCopied( event, db, name)
	self:Update()
	self:Print( format( L.ProfileCopied, name))
end

function Bison:OnProfileReset( event, db)
	self:Update()
	self:Print( format( L.ProfileReset, db:GetCurrentProfile()))
end

function Bison:OnProfileDeleted( event, db, name)
	self:Print( format( L.ProfileDeleted, name))
end

--[[---------------------------------------------------------------------------------
Layout
------------------------------------------------------------------------------------]]
AceGUI:RegisterLayout( 'BuffBar', function( content, children)
	local profile = Bison.db.profile[content.obj.moduleName]
	local xPadding = -profile.xPadding or 0
	local yPadding = -profile.yPadding or 0
	local horizontal = profile.horizontal
	local scale = profile.scale or 1
	local cols = profile.cols or 10
	local rows = profile.rows or 4
	local last, first
	local width = 0
	local height = 0
	local i = 1
	for k,child in pairs( children) do
		local frame = child.frame
		frame:ClearAllPoints()
		frame:SetScale( scale)
		if last and first then
			if horizontal then
				if xPadding >= 0 then
					frame:SetPoint( 'LEFT', last, 'RIGHT', xPadding, 0)
				else
					frame:SetPoint( 'RIGHT', last, 'LEFT', xPadding, 0)
				end
			else
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOM', last, 'TOP', 0, yPadding)
				else
					frame:SetPoint( 'TOP', last, 'BOTTOM', 0, yPadding)
				end
			end
		elseif first then
			if horizontal then
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOM', first, 'TOP', 0, yPadding)
				else
					frame:SetPoint( 'TOP', first, 'BOTTOM', 0, yPadding)
				end
			else
				if xPadding >= 0 then
					frame:SetPoint( 'LEFT', first, 'RIGHT', xPadding, 0)
				else
					frame:SetPoint( 'RIGHT', first, 'LEFT', xPadding, 0)
				end
			end
			first = frame
		else
			if xPadding >= 0 then
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOMLEFT', content)
				else
					frame:SetPoint( 'TOPLEFT', content)
				end
			else
				if yPadding >= 0 then
					frame:SetPoint( 'BOTTOMRIGHT', content)
				else
					frame:SetPoint( 'TOPRIGHT', content)
				end
			end
			first = frame
		end
		last = frame
		if child.DoLayout then
			child:DoLayout()
		end
		width  = max( width, frame.width or frame:GetWidth() or 0)
		height = max( height, (frame.height or frame:GetHeight() or 0) + (child.count and child.count:GetHeight() or 0))
		i = i + 1
		if horizontal then
			if i > cols then
				i = 1
				last = nil
			end
		else
			if i > rows then
				i = 1
				last = nil
			end
		end
	end
	content.obj:LayoutFinished( scale * cols * (width + abs(xPadding)), scale * rows * (height + abs(yPadding)))
end)
