--[[
Bison - a framework for free positioning of buffs.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:
	
Free Software Foundation, I.,
51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
--]]

local MAJOR = 'weapon'

local Bison  = LibStub( 'AceAddon-3.0'):GetAddon( 'Bison')
local L      = LibStub( 'AceLocale-3.0'):GetLocale( 'Bison')
local AceGUI = LibStub( 'AceGUI-3.0')

local UPDATE_TIME = 0.2
local MAX_BUTTON  = 2
local COLOR       = { r=0.2, g=0.2, b=0.2 }

local Module = Bison:NewBarModule( MAJOR)

local options = {
	type = 'group', order = 40, name = MAJOR, handler = Module, dialogHidden = true, dialogInline = true, get = 'GetProperty', set = 'SetProperty', 
	args = {
		option = { 
			type = 'group', order = 10, name = L.OptionName, inline = true, 
			args = {
				show     = { type = 'toggle', order = 10, name = L.ShowName,     desc = L.ShowDesc, width = 'full' },
				flashing = { type = 'toggle', order = 20, name = L.FlashingName, desc = L.FlashingDesc },
				timer    = { type = 'toggle', order = 30, name = L.TimerName,    desc = L.TimerDesc },
				sort     = { type = 'select', order = 40, name = L.SortName,     desc = L.SortDesc, get = 'GetSortType', set = 'SetSortType', values = 'GetSortDesc' },
			} 
		},
		layout = { 
			type = 'group', order = 20, name = L.BarName, inline = true, 
			args = {
				horizontal = { type = 'toggle', order = 10, name = L.HorizontalName, desc = L.HorizontalDesc },
				scale      = { type = 'range',  order = 20, name = L.ScaleName,      desc = L.ScaleDesc,    set = 'SetScale',    min = 0.01, max = 2,          step = 0.01, isPercent = true },
				cols       = { type = 'range',  order = 30, name = L.ColsName,       desc = L.ColsDesc,     set = 'SetCols',     min = 1,    max = MAX_BUTTON, step = 1 },
				rows       = { type = 'range',  order = 50, name = L.RowsName,       desc = L.RowsDesc,     set = 'SetRows',     min = 1,    max = MAX_BUTTON, step = 1 },
				xPadding   = { type = 'range',  order = 40, name = L.XPaddingName,   desc = L.XPaddingDesc, set = 'SetXPadding', min = -20,  max = 20,         step = 1 },
				yPadding   = { type = 'range',  order = 60, name = L.YPaddingName,   desc = L.YPaddingDesc, set = 'SetYPadding', min = -50,  max = 50,         step = 1 },
			} 
		}
	},
}

--[[---------------------------------------------------------------------------------
Script
------------------------------------------------------------------------------------]]
local function Button_OnEnter( button)
	GameTooltip:SetOwner( button.frame, 'ANCHOR_BOTTOMLEFT')
	GameTooltip:SetInventoryItem( PlayerFrame.unit, button.id + 15)
end

local function Button_OnLeave( button)
	if GameTooltip:IsOwned( button.frame) then
		GameTooltip:Hide()
	end
end

local function Button_OnMouseDown( button)
	Module.group:Fire( 'OnMouseDown')
end

local function Button_OnMouseUp( button)
	Module.group:Fire( 'OnMouseUp')
end

local function Button_OnTooltip( button)
	if GameTooltip:IsOwned( button.frame) then
		GameTooltip:SetInventoryItem( PlayerFrame.unit, button.id + 15)
	end
end

local function Bar_OnPosition( bar, event, x, y)
	Module.profile.xPos = x
	Module.profile.yPos = y
end

local function CreateButton( id)
	local btn = AceGUI:Create( 'BisonBuff')
	btn:Initialize( id, COLOR, FILTER)
	btn:SetCallback( 'OnEnter', Button_OnEnter)
	btn:SetCallback( 'OnLeave', Button_OnLeave)
	btn:SetCallback( 'OnMouseDown', Button_OnMouseDown)
	btn:SetCallback( 'OnMouseUp',   Button_OnMouseUp)
	btn:SetCallback( 'OnTooltip',   Button_OnTooltip)
	Bison:AddChieldToLBF( L.BarDebuff, btn)
	return btn
end

local function Bar_OnRenumber( bar)
	table.sort( bar.children, Module:GetSort( 'none'))
	local max = Module.profile.number
	while #bar.children > max do
		local child = tremove( bar.children)
		Bison:RemoveChieldFromLBF( L.BarWeapon, child)
		AceGUI:Release( child)
	end
	local n = #bar.children + 1
	for id = n,max do
		bar:AddChild( CreateButton( id))
	end
end

local function Bar_OnUpdate( bar, event, elapsed)
	for _,child in pairs( bar.children) do
		local user = child:GetUserDataTable()
		user.timeLeft = user.timeLeft and max( user.timeLeft - elapsed, 0)
	end
	if bar.update < 0 then
		bar.update = UPDATE_TIME
		Module:UpdateLeft()
	else
		bar.update = bar.update - elapsed
	end
end

--[[---------------------------------------------------------------------------------
Class
------------------------------------------------------------------------------------]]
function Module:OnInitialize()
	self.profile = Bison.db.profile[MAJOR]
	Bison:AddToLBF( L.BarWeapon, MAJOR)
	local group = self:CreateGroup()
	group.moduleName = MAJOR
	group:Initialize( COLOR, 0.06)
	group:SetCallback( 'OnPosition', Bar_OnPosition)
	group:SetCallback( 'OnRenumber', Bar_OnRenumber)
	group:SetCallback( 'OnUpdate',   Bar_OnUpdate)
	Bison:GetModule( 'Options'):AddPlugin( 'bison'..MAJOR, { [MAJOR] = options })
	LibStub( 'AceConfigDialog-3.0'):AddToBlizOptions( 'bison', L.BarWeapon, 'Bison', MAJOR)
end

function Module:OnEnable()
	Bison:MoveBlizzardFrame( self.group.frame)
	self:DoEnable()
end
	
function Module:OnDisable()
	self:DoDisable()
end

--[[---------------------------------------------------------------------------------
Main
------------------------------------------------------------------------------------]]
function Module:ReadValues()
	local mainEnchant, mainTimeLeft, mainCharges, offEnchant, offTimeLeft, offCharges = GetWeaponEnchantInfo()
	mainTimeLeft = mainEnchant and mainExpiration and (mainExpiration / 1000)
	offTimeLeft  = offEnchant and offExpiration and (offExpiration / 1000)
	local child1 = self.group.children[1]
	child1.id    = 16
	local user1 = child1:GetUserDataTable()
	user1.name     = mainEnchant 
	user1.count    = mainCharges 
	user1.texture  = mainEnchant and GetInventoryItemTexture( PlayerFrame.unit, 16) 
	user1.timeLeft = mainTimeLeft and (mainTimeLeft > 0) and mainTimeLeft or nil
	local child2 = self.group.children[2]
	child2.id    = 17
	local user2 = child2:GetUserDataTable()
	user2.name     = offEnchant 
	user2.count    = offCharges 
	user2.texture  = offEnchant and GetInventoryItemTexture( PlayerFrame.unit, 17) 
	user2.timeLeft = offTimeLeft and (offTimeLeft > 0) and offTimeLeft or nil
end

function Module:UpdateLeft()
	local timer = self.profile.timer
	local flashing = self.profile.flashing
	for _,child in pairs( self.group.children) do
		local user = child:GetUserDataTable()
		user.warn = user.timeLeft and (user.timeLeft < BUFF_DURATION_WARNING_TIME) or nil
		user.text = user.timeLeft and (SHOW_BUFF_DURATIONS == '1') and self:DurationString( user.timeLeft, timer) or nil
		child:UpdateLeft( flashing)
	end
end

